import abc
import os
from datetime import datetime, timedelta
from typing import List

import spotipy
from TheCodeLabs_BaseUtils.DefaultLogger import DefaultLogger

LOG_FORMAT = '[%(levelname)-7s] - %(asctime)s - %(message)s'
LOGGER = DefaultLogger().create_logger_if_not_exists('SpotifyBackup', logFormat=LOG_FORMAT)


class BackupHandler(abc.ABC):
    DATE_FORMAT = '%Y-%m-%d_%H-%M-%S'
    DATE_FORMAT_SHORT = '%Y-%m-%d'
    INVALID_CHARS = [' ', '\t', '\n']
    REPLACE_CHAR = '_'

    def __init__(self, exportFolder: str, clientID: str):
        self._clientID = clientID
        self._exportFolder = exportFolder

        if self._exportFolder:
            os.makedirs(self._exportFolder, exist_ok=True)

        self._spotify = self.login()

    @abc.abstractmethod
    def login(self) -> spotipy.Spotify:
        pass

    @abc.abstractmethod
    def backup(self):
        pass

    @abc.abstractmethod
    def _export_csv(self, exportPath: str, items: List):
        pass

    def _determine_export_path(self, fileName: str) -> str:
        for item in self.INVALID_CHARS:
            fileName = fileName.replace(item, self.REPLACE_CHAR)
        completeFileName = f'{datetime.strftime(datetime.now(), self.DATE_FORMAT)}_{fileName}.csv'
        exportPath = os.path.join(self._exportFolder, completeFileName)
        return exportPath

    def clean_old_exports(self, daysToKeep):
        if not daysToKeep:
            LOGGER.info('Skipping deletion of old files')
            return

        minimumDate = datetime.today() - timedelta(days=int(daysToKeep))
        LOGGER.info(f'>>> Deleting files older than {minimumDate} ({daysToKeep} days)')

        files = [file for file in sorted(os.listdir(self._exportFolder)) if
                 os.path.isfile(os.path.join(self._exportFolder, file))]

        for file in files:
            parts = file.split('_')
            creationDate = datetime.strptime(parts[0], self.DATE_FORMAT_SHORT)
            if creationDate < minimumDate:
                LOGGER.info(f'Removing old file "{file}"')
                os.remove(os.path.join(self._exportFolder, file))
