import csv
import json
from typing import Dict, List

import spotipy
from TheCodeLabs_BaseUtils.DefaultLogger import DefaultLogger
from spotipy import CacheFileHandler
from spotipy.oauth2 import SpotifyClientCredentials

from BackupHandler import BackupHandler

LOG_FORMAT = '[%(levelname)-7s] - %(asctime)s - %(message)s'
LOGGER = DefaultLogger().create_logger_if_not_exists('SpotifyBackup', logFormat=LOG_FORMAT)


class SpotifyBackup(BackupHandler):
    def __init__(self, clientID: str, clientSecret: str, exportFolder: str,
                 playlists: List[Dict[str, str]], cacheFilePath: str):
        self._clientSecret = clientSecret
        self._cacheFilePath = cacheFilePath
        super().__init__(exportFolder, clientID)
        self._playlists = playlists

    def login(self) -> spotipy.Spotify:
        client_credentials_manager = SpotifyClientCredentials(client_id=self._clientID,
                                                              client_secret=self._clientSecret,
                                                              cache_handler=CacheFileHandler(
                                                                  cache_path=self._cacheFilePath))
        return spotipy.Spotify(client_credentials_manager=client_credentials_manager)

    def backup(self):
        for playlist in SETTINGS['playlists']:
            self.__backup_playlist(playlist['user'], playlist['id'])

    def __backup_playlist(self, username: str, playlistID: str):
        playlist = self.__get_playlist(username, playlistID)
        tracks = self.__get_tracks(playlist)

        exportPath = self._determine_export_path(playlist['name'])

        LOGGER.info(f'>>> Exporting tracks to "{exportPath}"...')
        self._export_csv(exportPath, tracks)
        LOGGER.info('Exporting DONE')

    def __get_playlist(self, username: str, playlistID: str):
        LOGGER.info(f'>>> Fetching playlist with ID: {playlistID} by {username}...')
        identifier = f'spotify:user:{username}:playlist:{playlistID}'
        playlist = self._spotify.playlist(identifier)
        LOGGER.info(f'Found playlist "{playlist["name"]}"')
        return playlist

    def __get_tracks(self, playlist):
        tracks = playlist['tracks']
        results = tracks['items']
        while tracks['next']:
            tracks = self._spotify.next(tracks)
            results.extend(tracks['items'])

        LOGGER.info(f'Fetched {len(results)} tracks')
        return results

    def _export_csv(self, exportPath: str, items: List):
        with open(exportPath, 'w', newline='', encoding='utf-8') as f:
            writer = csv.writer(f, delimiter=';', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            writer.writerow(('title', 'artists', 'album', 'addedBy', 'addedAt', 'url'))

            for track in items:
                try:
                    trackInfo = track['track']
                    title = trackInfo['name']
                    artists = trackInfo['artists']
                    artists = ' & '.join(artist['name'] for artist in artists)
                    album = trackInfo['album']['name']
                    author = track['added_by']['id']
                    addedAt = track['added_at']
                    url = trackInfo['external_urls'].get('spotify', '')

                    writer.writerow((title, artists, album, author, addedAt, url))
                except Exception:
                    LOGGER.exception(f'Error while exporting track "{track}"')


if __name__ == '__main__':
    with open('config/settings.json', 'r', encoding='utf-8') as f:
        SETTINGS = json.load(f)

    spotifyBackup = SpotifyBackup(SETTINGS['spotifyAPI']['clientID'],
                                  SETTINGS['spotifyAPI']['clientSecret'],
                                  SETTINGS['exportFolder'],
                                  SETTINGS['playlists'],
                                  SETTINGS['cacheFilePath'])

    spotifyBackup.clean_old_exports(SETTINGS['daysToKeep'])
    spotifyBackup.backup()

    LOGGER.info('### DONE ###')
