import csv
import json
from typing import List

import spotipy
from TheCodeLabs_BaseUtils.DefaultLogger import DefaultLogger
from spotipy.oauth2 import SpotifyPKCE

from BackupHandler import BackupHandler

LOG_FORMAT = '[%(levelname)-7s] - %(asctime)s - %(message)s'
LOGGER = DefaultLogger().create_logger_if_not_exists('SpotifyFollowerBackup', logFormat=LOG_FORMAT)


class SpotifyFollowedArtistsBackup(BackupHandler):
    def __init__(self, clientID: str, exportFolder: str):
        super().__init__(exportFolder, clientID)

    def login(self) -> spotipy.Spotify:
        client_credentials_manager = SpotifyPKCE(client_id=self._clientID, redirect_uri='http://localhost:8080',
                                                 scope='user-follow-read')
        return spotipy.Spotify(client_credentials_manager=client_credentials_manager)

    def backup(self):
        followedArtists = self.__get_followed_artists()

        exportPath = self._determine_export_path('followed_artists')

        LOGGER.info(f'>>> Exporting followed artists to "{exportPath}"...')
        self._export_csv(exportPath, followedArtists)
        LOGGER.info('Exporting DONE')

    def __get_followed_artists(self):
        followedArtists = self._spotify.current_user_followed_artists(limit=20, after=None)['artists']

        results = followedArtists['items']
        while followedArtists['next']:
            followedArtists = self._spotify.next(followedArtists)['artists']
            results.extend(followedArtists['items'])

        LOGGER.info(f'Fetched {len(results)} followed artists')
        return results

    def _export_csv(self, exportPath: str, items: List):
        with open(exportPath, 'w', newline='', encoding='utf-8') as f:
            writer = csv.writer(f, delimiter=';', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            writer.writerow(('id', 'name', 'url', 'uri'))

            for artist in items:
                try:
                    id = artist['id']
                    name = artist['name']
                    url = artist['external_urls'].get('spotify', '')
                    uri = artist['uri']

                    writer.writerow((id, name, url, uri))
                except Exception:
                    LOGGER.exception(f'Error while exporting artist "{artist}"')


if __name__ == '__main__':
    with open('config/settings.json', 'r', encoding='utf-8') as f:
        SETTINGS = json.load(f)

    spotifyBackup = SpotifyFollowedArtistsBackup(SETTINGS['spotifyAPI']['clientID'], SETTINGS['exportFolder'])
    spotifyBackup.clean_old_exports(SETTINGS['daysToKeep'])
    spotifyBackup.backup()
    LOGGER.info('### DONE ###')
